;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-intermediate-lambda-reader.ss" "lang")((modname 09-3-paths) (read-case-sensitive #t) (teachpacks ()) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ())))
(require rackunit)
(require "extras.rkt")

(require "sets.rkt")

;; graph reachability 

;; In this example, we will refine our path? function to do
;; breadth-first and depth-first search.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; GRAPHS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; data definitions using list of edges:

;; A Node is a Symbol
;; A Graph is a ListOf<Edge> with no repeats

(define-struct edge (from to))
;; An Edge is a (make-edge Node Node)

(define graph1
  (list
    (make-edge 'a 'c)
    (make-edge 'b 'a)
    (make-edge 'b 'c)))

;; this is the graph from the slides
(define graph2
  (list
    (make-edge 'a 'b)
    (make-edge 'a 'c)
    (make-edge 'a 'd)
    (make-edge 'b 'c)
    (make-edge 'd 'c)
    (make-edge 'd 'f)
    (make-edge 'c 'e)
    (make-edge 'f 'g)
    (make-edge 'e 'g)
    (make-edge 'g 'b)))

(define (node=? n1 n2) (symbol=? n1 n2))

;; Node Graph -> Listof<Node>
(define (successors n1 loe)
  (map 
   edge-to  
   (filter
    (lambda (e) (node=? (edge-from e) n1))
    loe)))

(begin-for-test
  
  (check set-equal?
    (successors 'a (list (make-edge 'a 'b) (make-edge 'a 'c)))
    (list 'b 'c))
 
  (check set-equal?
    (successors 'a graph2)
    '(b c d)))

(begin-for-test

  (check-equal?
    (path? graph2 'a 'a) 
    true
    "there should be a path from a to a in graph2")

 (check-equal?
   (path? graph2 'a 'g) 
   true
   "there should be a path from a to g in graph2")

  (check-equal?
   (path? graph2 'b 'd)
   false
   "should find no path from b to d")
  
  (check-equal?
   (path? graph2 'd 'g)
   true
   "should find a path from d to g")
  
  (check-equal?
   (path? graph2 'e 'd)
   false
   "should find no path from e to d")
  
  (check-equal? (path? graph1 'a 'b) false)

  (check-equal? (path? graph1 'b 'c) true))

;; SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes
;; RETURNS: the set of all their immediate successors
(define (all-successors nodes graph)
  (foldr
    (lambda (node s)
      (set-union
        (successors node graph)
        s))
    empty
    nodes))

(begin-for-test
  (check set-equal?
    (all-successors empty graph2)
    empty)
  (check set-equal?
    (all-successors (list 'a 'c) graph2)
    (list 'b 'c 'd 'e))
  (check set-equal?
    (all-successors (list 'f 'g) graph2)
    (list 'g 'b)))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PATHS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; We'll start with the last version from 08-3

(define (path?.v0 graph src tgt)
  (local
    ((define (reachable-from? newest nodes)
       ;; RETURNS: true iff there is a path from src to tgt in graph
       ;; INVARIANT: newest is a subset of nodes
       ;; AND:
       ;;   (there is a path from src to tgt in graph)
       ;;   iff (there is a path from newest to tgt)
       ;; STRATEGY: general recursion
       ;; HALTING MEASURE: the number of graph nodes _not_ in 'nodes'
       (cond
         [(member tgt newest) true]
         [else (local
                 ((define candidates (set-diff 
                                       (all-successors newest graph)
                                       nodes)))
                 (cond
                   [(empty? candidates) false]
                   [else (reachable-from?
                           candidates
                           (append candidates nodes))]))])))
    (reachable-from? (list src) (list src))))

;;;;;;;;;;;;;;;; path?.v1 ;;;;;;;;;;;;;;;;

;; First, let's break out the local function, just to make it less
;; scary.  
;; We'll need to add tgt and graph to the arguments, so the function
;; will be self-contained

;; ListOfNodes ListOfNodes Node Graph -> Boolean
;; GIVEN: 
;; 1. The list 'nodes' of all the nodes we've seen
;; 2. The list of nodes whose successors we haven't taken
;; 3. The target node 'tgt' that we are trying to reach
;; 4. The graph we are searching
;; RETURNS: Is tgt reachable from any of the nodes in 'nodes'?
;; INVARIANT: newest is a subset of nodes
;; AND:
;;   (there is a path from src to tgt in graph)
;;   iff (there is a path from newest to tgt)
;; HALTING MEASURE: the number of graph nodes _not_ in 'nodes'

(define (reachable-from?.v1 newest nodes tgt graph)
  (cond
    [(member tgt newest) true]
    [else (local
            ((define candidates (set-diff 
                                  (all-successors newest graph)
                                  nodes)))
            (cond
              [(empty? candidates) false]
              [else (reachable-from?.v1
                      candidates
                      (append candidates nodes)
                      tgt
                      graph)]))]))

;; Strategy: Function Combination
(define (path?.v1 graph src tgt)
  (reachable-from?.v1 (list src) (list src) tgt graph))

;; All the action is in reachable-from?, so we'll refine that.

;;;;;;;;;;;;;;;; reachable-from-dfs? ;;;;;;;;;;;;;;;;

;; Refinement #1: instead of using all-successors, we'll look at the
;; successors of the nodes in newest one at a time.  We call this
;; reachable-from-dfs?, because this version searches depth-first.

(define (reachable-from-dfs? newest nodes tgt graph)
  (cond
    [(member tgt newest) true]
    [(empty? newest) false]
    [else (local
            ((define candidates (set-diff 
                                  (successors (first newest) graph)
                                  nodes)))
            (cond
              [(empty? candidates) 
               (reachable-from-dfs?
                 (rest newest)
                 nodes tgt graph)]
              [else (reachable-from-dfs?
                      (append candidates (rest newest))
                      (append candidates nodes)
                      tgt
                      graph)]))]))

(define (path-dfs? graph src tgt)
    (reachable-from-dfs? (list src) (list src) tgt graph))

;; this is depth-first search: it searches the newly-found nodes
;; first.

;;;;;;;;;;;;;;;; reachable-from-bfs? ;;;;;;;;;;;;;;;;

;; to get breadth-first, we just put the new candidates at the _end_
;; of newest:

(define (reachable-from-bfs? newest nodes tgt graph)
  (cond
    [(member tgt newest) true]
    [(empty? newest) false]
    [else (local
            ((define candidates (set-diff 
                                  (successors (first newest) graph)
                                  nodes)))
            (cond
              [(empty? candidates) 
               (reachable-from-dfs?
                 (rest newest)
                 nodes tgt graph)]
              [else (reachable-from-dfs?
                      (append (rest newest) candidates)
                      (append candidates nodes)
                      tgt
                      graph)]))]))

(define (path-bfs? graph src tgt)
    (reachable-from-bfs? (list src) (list src) tgt graph))



; "using path?.v0"  (define path? path?.v0) 
; "using path?.v1"  (define path? path?.v1) 
; "using path-dfs?" (define path? path-dfs?) 
"using path-bfs?" (define path? path-bfs?) 

(run-tests path-tests)


                         