;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-intermediate-lambda-reader.ss" "lang")((modname 09-3b-reachability) (read-case-sensitive #t) (teachpacks ()) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ())))
(require rackunit)
(require "extras.rkt")

(require "sets.rkt")

;; graph reachability 

;; data definitions using list of edges:

;; A Node is a Symbol
;; A Graph is a ListOfEdge with no repeats

(define-struct edge (from to))
;; An Edge is a (make-edge Node Node)

(define graph1
  (list
    (make-edge 'a 'c)
    (make-edge 'b 'a)
    (make-edge 'b 'c)))

;; this is the graph from the slides
(define graph2
  (list
    (make-edge 'a 'b)
    (make-edge 'a 'c)
    (make-edge 'a 'd)
    (make-edge 'b 'c)
    (make-edge 'd 'c)
    (make-edge 'd 'f)
    (make-edge 'c 'e)
    (make-edge 'f 'g)
    (make-edge 'e 'g)
    (make-edge 'g 'b)))

(define (node=? n1 n2) (symbol=? n1 n2))

;; Node Graph -> ListofNode
(define (successors n1 loe)
  (map 
   edge-to  
   (filter
    (lambda (e) (node=? (edge-from e) n1))
    loe)))

(begin-for-test
  (check set-equal?
    (successors 'a (list (make-edge 'a 'b) (make-edge 'a 'c)))
    (list 'b 'c))
  (check set-equal?
    (successors 'a graph2)
    '(b c d)))

;; SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes
;; RETURNS: the set of all their immediate successors
(define (all-successors nodes graph)
  (foldr
    (lambda (node s)
      (set-union
        (successors node graph)
        s))
    empty
    nodes))

(begin-for-test
  (check set-equal?
    (all-successors empty graph2)
    empty)
  (check set-equal?
    (all-successors (list 'a 'c) graph2)
    (list 'b 'c 'd 'e))
  (check set-equal?
    (all-successors (list 'f 'g) graph2)
    (list 'g 'b)))

#;(begin-for-test
  (check-equal?
   (path? graph2 'a 'a) 
   true
   "there should be a path from a to a in graph2")

  (check-equal?
   (path? graph2 'a 'g) 
   true
   "there should be a path from a to g in graph2")

  (check-equal?
   (path? graph2 'b 'd)
   false
   "should find no path from b to d")
  
  (check-equal?
   (path? graph2 'd 'g)
   true
   "should find a path from d to g")
  
  (check-equal?
   (path? graph2 'e 'd)
   false
   "should find no path from e to d")
  
  (check-equal? (path? graph1 'a 'b) false)

  (check-equal? (path? graph1 'b 'c) true))

  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; reachables : SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes in a graph
;; RETURNS: the set of nodes reachable from 'nodes'

;; EXAMPLES/TESTS
(begin-for-test
  (check set-equal?
    (reachables (list 'b) graph2)
    (list 'b 'c 'e 'g))
  (check set-equal?
    (reachables (list 'a) graph2)
    (list 'a 'b 'c 'd 'e 'f 'g)))

;; this is a highly inefficient way to define path?
(define (path? src tgt g)
  (member tgt (reachables (list src) g)))



;;; reachables: five versions

;; reachables : SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes in a graph
;; RETURNS: the set of nodes reachable from 'nodes'

;;;;;;;;;;;;;;;; reachables.v1 ;;;;;;;;;;;;;;;;

;; reachables.v1 : SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes in a graph
;; RETURNS: the set of nodes reachable from 'nodes'
;; STRATEGY: General Recursion
;; HALTING MEASURE: the set of graph nodes NOT in nodes
(define (reachables.v1 nodes graph)
  (local
    ((define candidates (all-successors nodes graph)))
    (cond
      [(subset? candidates nodes) nodes]
      [else (reachables.v1
              (set-union candidates nodes)
              graph)])))

;; TERMINATION ARGUMENT: At the recursive call, 'candidates' contains at
;; least one element that is not in 'nodes' (otherwise the subset? test
;; would have returned true).  Hence the result of the set-union is at
;; least one element bigger than 'nodes'.  So the halting measure
;; decreases. 

;; This is called a CLOSURE ALGORITHM: we want to find the smallest
;; set containing nodes and which is closed under successors.

;; This version keeps looking at the successors of the original
;; nodes.  We only need to look at the successors of the most recently
;; added nodes. We'll do that in reachables.v2.

;;;;;;;;;;;;;;;; reachables.v2 ;;;;;;;;;;;;;;;;

;; reachable-from.v2 : ListOfNode ListOfNode Graph -> SetOfNode
;; GIVEN: two lists of nodes: 'worklist' and 'explored'
;; WHERE: the two lists are disjoint
;; AND: (reachables (worklist U explored)) = (reachables worklist)
;; AND: (successors explored) subset (worklist U explored)
;; RETURNS: the set of nodes reachable from (worklist U explored)
;; STRATEGY: general recursion
;; HALTING MEASURE: the number of graph nodes _not_ in explored
(define (reachable-from.v2 worklist explored graph)
  (cond
    [(empty? worklist) explored]
    [else (local
            ((define candidates (set-diff
                                  (all-successors worklist graph)
                                  explored)))
            (reachable-from.v2
              candidates
              (append worklist explored)
              graph))]))

;; CORRECTNESS:
;; if worklist is empty, then
;; (successors explored) \subset (worklist U explored) = explored.
;; So (reachables explored) = explored
;; Hence reachables(worklist U explored)
;;       = reachables(explored)
;;       = explored

;; TERMINATION ARGUMENT:
;; At the recursive call, worklist is non-empty (because of the test)
;; and is disjoint from explored (because of the invariant).  
;; Hence the new value of (worklists U explored) is at least one
;; element larger than the old one.  And therefore the halting measure
;; decreases. 

;; This is called the "worklist algorithm". We apply some function to
;; the worklist and see if it discovers any new nodes,

;; we initialize the worklist to nodes and the explored to empty.
;; This makes the invariant true.

(define (reachables.v2 nodes graph)
  (reachable-from.v2 nodes empty graph))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; v3: breadth-first search

;; for our next refinement, we're going to stop saying (all-successors
;; worklist), and process the elements of the worklist one at a time.

;; The simplest version of this is BREADTH-FIRST SEARCH.

;; In this version, we will allow worklist to have duplicates.

;; reachable-from.v3 : ListOfNode ListOfNode Graph -> ListOfNode
;; GIVEN: two lists of nodes: 'worklist' and 'explored'
;; WHERE: the two lists are disjoint
;; AND: (reachables (worklist U explored)) = (reachables worklist)
;; AND: (successors explored) subset (worklist U explored)
;; RETURNS: the set of nodes reachable from (worklist U explored)
;; STRATEGY: general recursion
;; HALTING MEASURE: the number of graph nodes _not_ in explored
(define (reachable-from.v3 worklist explored graph)
  (cond
    [(empty? worklist) explored]
    [else (local
            ((define candidates (set-diff
                                  (successors (first worklist) graph)
                                  explored)))
            (reachable-from.v3
              (append (rest worklist) candidates)
              (cons (first worklist) explored)
              graph))]))

;; we initialize the invariant in the same way:

(define (reachables.v3 nodes graph)
  (reachable-from.v3 nodes empty graph))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; v4: almost depth-first search

;; In breadth-first search, the worklist was organized as a queue.
;; In depth-first search, the worklist is organized as a stack. 

;; This isn't quite depth-first search, but it's close
;; All the invariants are the same:

(define (reachable-from.v4 worklist explored graph)
  (cond
    [(empty? worklist) explored]
    [else (local
            ((define candidates (set-diff
                                  (successors (first worklist) graph)
                                  explored)))
            (reachable-from.v4
              (append candidates (rest worklist))
              (cons (first worklist) explored)
              graph))]))


;; we initialize the invariant in the same way:

(define (reachables.v4 nodes graph)
  (reachable-from.v4 nodes empty graph))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; v5: depth-first search

;;; to get real depth-first search, we abandon the constraint that
;;; worklist and explored are disjoint.  If a node shows up on the
;;; worklist that is already explored, we simply discard it.

;;; We can also abandon the constraint that worklist has no duplicates.

;; This means that EITHER the worklist gets shorter or the explored
;; list gets longer (and hence the set of unexplored nodes gets smaller).

;;; So the halting measure is the pair

;;; (# of unexplored nodes, length of worklist)

;;; under the lexicographic ordering:  the length of the worklist
;;; cannot decrease indefinitely without decreasing the number of
;;; unexplored nodes.

(define (reachable-from.v5 worklist explored graph)
  (cond
    [(empty? worklist) explored]
    [(member? (first worklist) explored)
     (reachable-from.v5
       (rest worklist)
       explored
       graph)]
    [else
      (local
        ((define candidates (successors (first worklist) graph)))
        (reachable-from.v5
          (append candidates (rest worklist))
          (cons (first worklist) explored)
          graph))]))

;; we initialize the invariant in the same way:

(define (reachables.v5 nodes graph)
  (reachable-from.v5 nodes empty graph))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; provided by sets.rkt
#;(define (set-diff set1 set2)
  (filter
    (lambda (x) (not (my-member? x set2)))
    set1))

;; comment out all but one of these lines:
;; "v1" (define reachables reachables.v1)
;; "v2" (define reachables reachables.v2)
;; "v3" (define reachables reachables.v3)
;; "v4" (define reachables reachables.v4)
 "v5" (define reachables reachables.v5)
