;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-intermediate-lambda-reader.ss" "lang")((modname 09-4-queens) (read-case-sensitive #t) (teachpacks ()) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ())))
;; N-queens problem

(require rackunit)
(require "extras.rkt")
(require "sets.rkt")

;; A classic problem:

;; Place N queens on an NxN chessboard such that no two queens
;; threaten each other.

;; Queens at (r1,c1) and (r2,c2) threaten each other iff either
;; r1=r2, c1=c2, r1+c1 = r2+c2 or r1-c1 = r2-c2

;; We will model this as a graph-search problem.

;; Nodes of the graph:  legal configurations
;; {(1,c1),(2,c2),...} 
;; We'll fill the board one row at a time
;; Given a legal configuration 
;; config = {(1,c1),(2,c2),...(k, c_k)}
;; the successors of config are those legal configurations
;; {(1,c1),(2,c2),...(k, c_k), (k+1, c_k+1)}

;; Data Design
;; The data design should accomodate boards of any size.


;; Queens:
(define-struct queen (row col))
;; A Queen is a (make-queen PosInt PosInt)

;; Queen Queen -> Boolean
(define (threatens? q1 q2)
  (or
    (= (queen-row q1) (queen-row q2))
    (= (queen-col q1) (queen-col q2))
    (= 
      (+ (queen-row q1) (queen-col q1))
      (+ (queen-row q2) (queen-col q2)))
    (= 
      (- (queen-row q1) (queen-col q1))
      (- (queen-row q2) (queen-col q2)))))

;; Queen ListOf<Queens> -> Boolean
(define (threatens-any? this-queen other-queens)
  (ormap
    (lambda (other-queen) (threatens? this-queen other-queen))
    other-queens))


;; Configurations:
;; A Config is a ListOf<Queen>
;; WHERE: the queens are listed in decreasing row order, eg
;; ((k, c_k), (k-1, c_k-1), ... (1, c1))
;; AND WHERE: no two of the queens threaten each other.

;; : -> Config
(define empty-config empty)

;; cons-config : PosInt Config -> Maybe<Config>
;; GIVEN: a column col and a config ((k, c_k), (k-1, c_k-1), ... (1,
;; c1))
;; RETURNS: the configuration ((k+1, col) (k, c_k), (k-1, c_k-1),
;; ... (1, c1)) if that is a legal configuration, otherwise false.
;; Strategy: function combination
;; Algorithm: Check to see if the new queen threatens any of the
;; existing queens.
(define (cons-config col config)
  (if (empty? config)
    ;; if the configuration is empty, then the first move is in row 1
    ;; and is always legal
    (list (make-queen 1 col))
    (local
      ((define next-row (+ 1 (queen-row (first config))))
       (define new-queen (make-queen next-row col)))
      (if
        (threatens-any? new-queen config)
        false
        (cons new-queen config)))))

;; RETURNS: Is the configuration complete for a board of size n?
(define (complete? size config)
  (= size (length config)))

;; RETURNS: Is the configuration legal?
(define (config-legal? config)
  (cond
    [(empty? config) true]
    [else (and
            (not (threatens-any? (first config) (rest config)))
            (config-legal? (rest config)))]))

(begin-for-test
  "config-legal tests; "
  (check-true (config-legal? empty))
  (check-false (config-legal? (list (make-queen 1 2) (make-queen 1 3))))
  (check-false (config-legal? (list (make-queen 1 2) (make-queen 7 8) 
                                (make-queen 3 5))))
  (check-true (config-legal? 
                (list (make-queen 4 3) (make-queen 3 1) 
                  (make-queen 2 4)
                  (make-queen 1 2)))))



;; Config PosInt -> ListOf<Config>
;; GIVEN: a configuration config and a board size
;; RETURNS: the set of all the successors of config with columns in
;; [1,size]
;; STRATEGY: HOFC
(define (successors config size)
  (filter
    (lambda (v) (not (false? v)))
    (map
     (lambda (col) (cons-config col config))
     (build-columns size))))

;; PosInt -> ListOf<PosInt>
;; GIVEN: PosInt n
;; RETURNS: (list 1 .. n)       
(define (build-columns n)
  (build-list n (lambda (i) (+ i 1))))

;; SetOfNode Graph -> SetOfNode
;; GIVEN: A set of nodes
;; RETURNS: the set of all their immediate successors
;; since there are no duplicates, we can use append instead of set-union
(define (all-successors nodes size)
  (foldr
    (lambda (node s)
      (append
        (successors node size)
        s))
    empty
    nodes))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; OK, now we're ready use the graph-search program.
;;; We'll start with the final path? program.

;;; As before, we'll break out reachable-from.

;;;;;;;;;;;;;;;; placement-reachable-from?.v1 ;;;;;;;;;;;;;;;;

;;; In place of 'graph' we'll use the size of the board.
;;; In place of src, we'll use the empty configuration
;;; In place of tgt, we'll just test to see if we ever get a
;;; configuration whose length is size.

;;; And instead of returning a Boolean, we'll return the successful
;;; configuration. 

;; placement-reachable-from?: ListOfNodes ListOfNodes PosInt -> Maybe<Config>
;; GIVEN: 
;; 1. The list of nodes 'newest' whose successors we haven't taken
;; 2. The list 'nodes' of all the nodes we've seen
;; 3. The size of the chessboard
;; RETURNS: a complete configuration of the queens, if one exists,
;; otherwise false. 
;; INVARIANT: newest is a subset of nodes
;; AND:
;;   (there is a some complete placement of queens)
;;   iff (there is some complete placement that extends one of the
;;        nodes in 'newest')
;; HALTING MEASURE: the number of graph nodes _not_ in 'nodes'

(define (placement-reachable-from?.v1 newest nodes size)
  (cond
    [(empty? newest) false]
    [(any-complete? size newest)
     (first-complete size newest)]
    [else (local
            ((define candidates (set-diff 
                                  (successors (first newest) size)
                                  nodes)))
            (cond
              [(empty? candidates) 
               (placement-reachable-from?.v1 (rest newest) nodes size)]
              [else (placement-reachable-from?.v1
                      (append candidates (rest newest))
                      (append candidates nodes)
                      size)]))]))

(define (queens-dfs.v1 size)
  (placement-reachable-from?.v1
    (list empty-config) (list empty-config) size))

;; PosInt ListOf<Config> -> Boolean
(define (any-complete? size configs)
  (ormap
    (lambda (c) (complete? size c))
    configs))

;; PosInt ListOf<Config> -> Config
;; WHERE: one of the configs is guaranteed to be complete.
(define (first-complete size configs)
  (first
    (filter
      (lambda (c) (complete? size c))
      configs)))

;;;;;;;;;;;;;;;; placement-reachable-from?.v2 ;;;;;;;;;;;;;;;;

;;; Next version:  observe that successor generates a node we've seen
;;; before.  So we don't need the set-diff:

(define (placement-reachable-from?.v2 newest nodes size)
  (cond
    [(empty? newest) false]
    [(any-complete? size newest)
     (first-complete size newest)]
    [else (local
            ((define candidates (successors (first newest) size)))
            (cond
              [(empty? candidates) 
               (placement-reachable-from?.v2 (rest newest) nodes size)]
              [else (placement-reachable-from?.v2
                      (append candidates (rest newest))
                      (append candidates nodes)
                      size)]))]))

(define (queens-dfs.v2 size)
  (placement-reachable-from?.v2
    (list empty-config) (list empty-config) size))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; We've used this to illustrate an application of graph search, but
;;; we can get a simpler program by returning to first principles and
;;; applying general recursion directly.

;;; The problem is to find some completion of the empty configuration.
;;; So the obvious thing is to generalize away from the empty configuration:

;;; A different approach is "depth-first" search:  given a
;;; configuration c, is there _any_ completion?

;; if c is already complete, it is its own completion.
;; otherwise, look at each of the successors of c in turn, and choose
;; the first completion

;; generalized-queens-dfs : Config PosInt-> Maybe<Config>
;; GIVEN: a configuration
;; RETURNS: a completion of the configuration, if there is one,
;; otherwise false.
;; STRATEGY: general recursion
;; HALTING MEASURE: (- size (length c))
(define (generalized-queens-dfs c size)
  (if (complete? size c)
    c
    (first-success
      (lambda (new-c) (generalized-queens-dfs new-c size))
      (successors c size))))

;; first-success is similar to ormap, but it returns the first
;; non-false value

;; (X -> Maybe<Y>) ListOf<X> -> Maybe<Y>
;; GIVEN: lst = (list x_1 x_2 ...)
;; RETURNS: The first non-false value of (f x_1), (f x_2), etc.
(define (first-success fn lst)
  (cond
    [(empty? lst) false]
    [else (local
            ((define firstval (fn (first lst))))
            (if (false? firstval)
              (first-success fn (rest lst))
              firstval))]))

(define (queens-dfs.v3 size)
  (generalized-queens-dfs empty-config size))

(define (test-all n)
  (list
    (queens-dfs.v1 n)
    (queens-dfs.v2 n)
    (queens-dfs.v3 n)))

(define (all-legal? configs)
  (andmap config-legal? configs))

(begin-for-test
  (check-true (all-legal? (test-all 5)))
  (check-true (all-legal? (test-all 4)))
  (check-true (all-legal? (test-all 8)))
  (check-true (all-legal? (test-all 10))))



